@@ -73,9 +73,8 @@ def fit(model, x, y, get=threaded.get, **kwargs):
     name = 'fit-' + tokenize(model, x, y, kwargs)
     dsk = {(name, -1): model}
     dsk.update(dict(((name, i), (_partial_fit, (name, i - 1),
-                                          (x.name, i, 0),
-                                          (y.name, i),
-                                          kwargs))
+                                               (x.name, i, 0),
+                                               (y.name, i), kwargs))
                     for i in range(nblocks)))
 
     return get(merge(x.dask, y.dask, dsk), (name, nblocks - 1))
