@@ -3,15 +3,18 @@ from __future__ import absolute_import, division, print_function
 from itertools import product
 from math import ceil
 from numbers import Number
-from operator import getitem, add
+from operator import getitem, add, itemgetter
 
 import numpy as np
-from toolz import merge, first, accumulate, pluck
+from toolz import merge, accumulate, pluck, memoize
 
 from ..base import tokenize
 from ..compatibility import long
 
 
+colon = slice(None, None, None)
+
+
 def sanitize_index(ind):
     """ Sanitize the elements for indexing along one axis
 
@@ -148,23 +151,35 @@ def slice_with_newaxes(out_name, in_name, blockdims, index):
     # Strip Nones from index
     index2 = tuple([ind for ind in index if ind is not None])
     where_none = [i for i, ind in enumerate(index) if ind is None]
+    where_none_orig = list(where_none)
+    for i, x in enumerate(where_none):
+        n = sum(isinstance(ind, int) for ind in index[:x])
+        if n:
+            where_none[i] -= n
 
     # Pass down and do work
     dsk, blockdims2 = slice_wrap_lists(out_name, in_name, blockdims, index2)
 
-    # Insert ",0" into the key:  ('x', 2, 3) -> ('x', 0, 2, 0, 3)
-    dsk2 = dict(((out_name,) + insert_many(k[1:], where_none, 0),
-                 (v[:2] + (insert_many(v[2], where_none, None),)))
+    if where_none:
+        expand = expander(where_none)
+        expand_orig = expander(where_none_orig)
+
+        # Insert ",0" into the key:  ('x', 2, 3) -> ('x', 0, 2, 0, 3)
+        dsk2 = {(out_name,) + expand(k[1:], 0):
+                (v[:2] + (expand_orig(v[2], None),))
                 for k, v in dsk.items()
-                if k[0] == out_name)
+                if k[0] == out_name}
 
-    # Add back intermediate parts of the dask that weren't the output
-    dsk3 = merge(dsk2, dict((k, v) for k, v in dsk.items() if k[0] != out_name))
+        # Add back intermediate parts of the dask that weren't the output
+        dsk3 = merge(dsk2, {k: v for k, v in dsk.items() if k[0] != out_name})
 
-    # Insert (1,) into blockdims:  ((2, 2), (3, 3)) -> ((2, 2), (1,), (3, 3))
-    blockdims3 = insert_many(blockdims2, where_none, (1,))
+        # Insert (1,) into blockdims:  ((2, 2), (3, 3)) -> ((2, 2), (1,), (3, 3))
+        blockdims3 = expand(blockdims2, (1,))
 
-    return dsk3, blockdims3
+        return dsk3, blockdims3
+
+    else:
+        return dsk, blockdims2
 
 
 def slice_wrap_lists(out_name, in_name, blockdims, index):
@@ -181,7 +196,8 @@ def slice_wrap_lists(out_name, in_name, blockdims, index):
     """
     shape = tuple(map(sum, blockdims))
     assert all(isinstance(i, (slice, list, int, long)) for i in index)
-    assert len(blockdims) == len(index)
+    if not len(blockdims) == len(index):
+        raise IndexError("Too many indices for array")
     for bd, i in zip(blockdims, index):
         check_index(i, sum(bd))
 
@@ -193,15 +209,14 @@ def slice_wrap_lists(out_name, in_name, blockdims, index):
     if len(where_list) > 1:
         raise NotImplementedError("Don't yet support nd fancy indexing")
 
+    # No lists, hooray! just use slice_slices_and_integers
+    if not where_list:
+        return slice_slices_and_integers(out_name, in_name, blockdims, index2)
+
     # Replace all lists with full slices  [3, 1, 0] -> slice(None, None, None)
     index_without_list = tuple(slice(None, None, None)
-                                    if isinstance(i, list)
-                                    else i
-                                    for i in index2)
-
-    # No lists, hooray! just use slice_slices_and_integers
-    if index2 == index_without_list:
-        return slice_slices_and_integers(out_name, in_name, blockdims, index2)
+                               if isinstance(i, list) else i
+                               for i in index2)
 
     # lists and full slices.  Just use take
     if all(isinstance(i, list) or i == slice(None, None, None)
@@ -252,14 +267,14 @@ def slice_slices_and_integers(out_name, in_name, blockdims, index):
     # (out_name, 0, 0, 0), (out_name, 0, 0, 1), (out_name, 0, 1, 0), ...
     out_names = list(product([out_name],
                              *[range(len(d))[::-1] if i.step and i.step < 0 else range(len(d))
-                                 for d, i in zip(block_slices, index)
-                                 if not isinstance(i, (int, long))]))
+                               for d, i in zip(block_slices, index)
+                               if not isinstance(i, (int, long))]))
 
     all_slices = list(product(*[pluck(1, s) for s in sorted_block_slices]))
 
-    dsk_out = dict((out_name, (getitem, in_name, slices))
-                   for out_name, in_name, slices
-                   in zip(out_names, in_names, all_slices))
+    dsk_out = {out_name: (getitem, in_name, slices)
+               for out_name, in_name, slices
+               in zip(out_names, in_names, all_slices)}
 
     new_blockdims = [new_blockdim(d, db, i)
                      for d, i, db in zip(shape, index, blockdims)
@@ -294,6 +309,11 @@ def _slice_1d(dim_shape, lengths, index):
     Examples
     --------
 
+    Trivial slicing
+
+    >>> _slice_1d(100, [60, 40], slice(None, None, None))
+    {0: slice(None, None, None), 1: slice(None, None, None)}
+
     100 length array cut into length 20 pieces, slice 0:35
 
     >>> _slice_1d(100, [20, 20, 20, 20, 20], slice(0, 35))
@@ -339,6 +359,9 @@ def _slice_1d(dim_shape, lengths, index):
 
     assert isinstance(index, slice)
 
+    if index == colon:
+        return {k: colon for k in range(len(lengths))}
+
     step = index.step or 1
     if step > 0:
         start = index.start or 0
@@ -401,20 +424,13 @@ def partition_by_size(sizes, seq):
     >>> partition_by_size([10, 20, 10], [1, 5, 9, 12, 29, 35])
     [[1, 5, 9], [2, 19], [5]]
     """
-    seq = list(seq)
-    pretotal = 0
-    total = 0
-    i = 0
-    result = list()
-    for s in sizes:
-        total += s
-        L = list()
-        while i < len(seq) and seq[i] < total:
-            L.append(seq[i] - pretotal)
-            i += 1
-        result.append(L)
-        pretotal += s
-    return result
+    seq = np.array(seq)
+    right = np.cumsum(sizes)
+    locations = np.searchsorted(seq, right)
+    locations = [0] + locations.tolist()
+    left = [0] + right.tolist()
+    return [(seq[locations[i]:locations[i + 1]] - left[i]).tolist()
+            for i in range(len(locations) - 1)]
 
 
 def issorted(seq):
@@ -435,9 +451,6 @@ def issorted(seq):
     return True
 
 
-colon = slice(None, None, None)
-
-
 def take_sorted(outname, inname, blockdims, index, axis=0):
     """ Index array with sorted list index
 
@@ -472,7 +485,7 @@ def take_sorted(outname, inname, blockdims, index, axis=0):
 
     outdims = list(dims)
     outdims[axis] = where_index
-    slices = [[colon]*len(bd) for bd in blockdims]
+    slices = [[colon] * len(bd) for bd in blockdims]
     slices[axis] = index_lists
     slices = list(product(*slices))
     inkeys = list(product([inname], *outdims))
@@ -518,21 +531,19 @@ def take(outname, inname, blockdims, index, axis=0):
     index_lists = partition_by_size(sizes, sorted(index))
 
     dims = [[0] if axis == i else list(range(len(bd)))
-                for i, bd in enumerate(blockdims)]
+            for i, bd in enumerate(blockdims)]
     keys = list(product([outname], *dims))
 
     rev_index = list(map(sorted(index).index, index))
     vals = [(getitem, (np.concatenate,
-                  (list, [(getitem, ((inname,) + d[:axis] + (i,) + d[axis+1:]),
-                                   ((colon,)*axis + (IL,) + (colon,)*(n-axis-1)))
-                                for i, IL in enumerate(index_lists)
-                                if IL]),
-                        axis),
-                     ((colon,)*axis + (rev_index,) + (colon,)*(n-axis-1)))
+                       [(getitem, ((inname, ) + d[:axis] + (i, ) + d[axis + 1:]),
+                         ((colon, ) * axis + (IL, ) + (colon, ) * (n - axis - 1)))
+                        for i, IL in enumerate(index_lists) if IL], axis),
+             ((colon, ) * axis + (rev_index, ) + (colon, ) * (n - axis - 1)))
             for d in product(*dims)]
 
     blockdims2 = list(blockdims)
-    blockdims2[axis] = (len(index),)
+    blockdims2[axis] = (len(index), )
 
     return tuple(blockdims2), dict(zip(keys, vals))
 
@@ -560,7 +571,7 @@ def posify_index(shape, ind):
         else:
             return ind
     if isinstance(ind, list):
-        return [posify_index(shape, i) for i in ind]
+        return [i + shape if i < 0 else i for i in ind]
     return ind
 
 
@@ -580,6 +591,42 @@ def insert_many(seq, where, val):
     return tuple(result)
 
 
+@memoize
+def _expander(where):
+    if not where:
+        def expand(seq, val):
+            return seq
+        return expand
+    else:
+        decl = """def expand(seq, val):
+            return ({left}) + tuple({right})
+        """
+        left = []
+        j = 0
+        for i in range(max(where) + 1):
+            if i in where:
+                left.append("val, ")
+            else:
+                left.append("seq[%d], " % j)
+                j += 1
+        right = "seq[%d:]" % j
+        left = "".join(left)
+        decl = decl.format(**locals())
+        ns = {}
+        exec(compile(decl, "<dynamic>", "exec"), ns, ns)
+        return ns['expand']
+
+
+def expander(where):
+    """ An optimized version of insert_many() when *where*
+    is known upfront and used many times.
+
+    >>> expander([0, 2])(['a', 'b', 'c'], 'z')
+    ('z', 'a', 'z', 'b', 'c')
+    """
+    return _expander(tuple(where))
+
+
 def new_blockdim(dim_shape, lengths, index):
     """
 
@@ -595,9 +642,10 @@ def new_blockdim(dim_shape, lengths, index):
     if isinstance(index, list):
         return [len(index)]
     assert not isinstance(index, (int, long))
-    pairs = sorted(_slice_1d(dim_shape, lengths, index).items(), key=first)
+    pairs = sorted(_slice_1d(dim_shape, lengths, index).items(),
+                   key=itemgetter(0))
     slices = [slice(0, lengths[i], 1) if slc == slice(None, None, None) else slc
-                for i, slc in pairs]
+              for i, slc in pairs]
     if isinstance(index, slice) and index.step and index.step < 0:
         slices = slices[::-1]
     return [int(ceil((1. * slc.stop - slc.start) / slc.step)) for slc in slices]
@@ -608,16 +656,19 @@ def replace_ellipsis(n, index):
 
     >>> replace_ellipsis(4, (3, Ellipsis, 2))
     (3, slice(None, None, None), slice(None, None, None), 2)
+
+    >>> replace_ellipsis(2, (Ellipsis, None))
+    (slice(None, None, None), slice(None, None, None), None)
     """
     # Careful about using in or index because index may contain arrays
     isellipsis = [i for i, ind in enumerate(index) if ind is Ellipsis]
+    extra_dimensions = n - (len(index) - sum(i is None for i in index) - 1)
     if not isellipsis:
         return index
     else:
         loc = isellipsis[0]
-    return (index[:loc]
-          + (slice(None, None, None),) * (n - len(index) + 1)
-          + index[loc+1:])
+    return (index[:loc] + (slice(None, None, None),) * extra_dimensions +
+            index[loc + 1:])
 
 
 def check_index(ind, dimension):
@@ -646,21 +697,21 @@ def check_index(ind, dimension):
     >>> check_index([6, 3], 5)
     Traceback (most recent call last):
     ...
-    IndexError: Index is not smaller than dimension 6 >= 5
+    IndexError: Index out of bounds 5
 
     >>> check_index(slice(0, 3), 5)
     """
     if isinstance(ind, list):
-        for i in ind:
-            check_index(i, dimension)
+        x = np.array(ind)
+        if (x >= dimension).any() or (x <= -dimension).any():
+            raise IndexError("Index out of bounds %s" % dimension)
     elif isinstance(ind, slice):
         return
 
     elif ind >= dimension:
-        raise IndexError("Index is not smaller than dimension %d >= %d"
-                % (ind, dimension))
+        raise IndexError("Index is not smaller than dimension %d >= %d" %
+                         (ind, dimension))
 
-    elif ind <= -dimension:
-        raise IndexError(
-                "Negative index is not greater than negative dimension %d <= -%d"
-                % (ind, dimension))
+    elif ind < -dimension:
+        msg = "Negative index is not greater than negative dimension %d <= -%d"
+        raise IndexError(msg % (ind, dimension))
