@@ -8,6 +8,7 @@ import time
 from distutils.version import LooseVersion
 from operator import add, sub, getitem
 from threading import Lock
+import warnings
 
 from toolz import merge, countby, concat
 from toolz.curried import identity
@@ -17,19 +18,20 @@ import dask.array as da
 from dask.base import tokenize
 from dask.delayed import delayed
 from dask.async import get_sync
-from dask.utils import raises, ignoring, tmpfile, tmpdir
+from dask.utils import ignoring, tmpfile, tmpdir
 from dask.utils_test import inc
 
 from dask.array import chunk
 from dask.array.core import (getem, getarray, getarray_nofancy, top, dotmany,
                              concatenate3, broadcast_dimensions, Array, stack,
                              concatenate, from_array, take, elemwise, isnull,
-                             notnull, broadcast_shapes, partial_by_order, exp,
+                             notnull, broadcast_shapes, partial_by_order,
                              tensordot, choose, where, coarsen, insert,
                              broadcast_to, reshape, fromfunction,
                              blockdims_from_blockshape, store, optimize,
                              from_func, normalize_chunks, broadcast_chunks,
-                             atop, from_delayed, concatenate_axes)
+                             atop, from_delayed, concatenate_axes,
+                             common_blockdim)
 from dask.array.utils import assert_eq
 
 # temporary until numpy functions migrated
@@ -144,6 +146,15 @@ def test_transpose():
               x.transpose((2, 0, 1)))
     assert same_keys(d.transpose((2, 0, 1)), d.transpose((2, 0, 1)))
 
+    assert_eq(d.transpose(2, 0, 1),
+              x.transpose(2, 0, 1))
+    assert same_keys(d.transpose(2, 0, 1), d.transpose(2, 0, 1))
+
+    with pytest.raises(ValueError):
+        d.transpose(1, 2)
+    with pytest.raises(ValueError):
+        d.transpose((1, 2))
+
 
 def test_broadcast_dimensions_works_with_singleton_dimensions():
     argpairs = [('x', 'i')]
@@ -167,7 +178,7 @@ def test_Array():
     assert a.numblocks == (10, 10)
 
     assert a._keys() == [[('x', i, j) for j in range(10)]
-                                     for i in range(10)]
+                         for i in range(10)]
 
     assert a.chunks == ((100,) * 10, (100,) * 10)
 
@@ -195,7 +206,7 @@ def test_keys():
     dsk = dict((('x', i, j), ()) for i in range(5) for j in range(6))
     dx = Array(dsk, 'x', chunks=(10, 10), shape=(50, 60))
     assert dx._keys() == [[(dx.name, i, j) for j in range(6)]
-                                          for i in range(5)]
+                          for i in range(5)]
     d = Array({}, 'x', (), shape=())
     assert d._keys() == [('x',)]
 
@@ -210,7 +221,7 @@ def test_Array_computation():
 def test_stack():
     a, b, c = [Array(getem(name, chunks=(2, 3), shape=(4, 6)),
                      name, shape=(4, 6), chunks=(2, 3))
-                for name in 'ABC']
+               for name in 'ABC']
 
     s = stack([a, b, c], axis=0)
 
@@ -219,35 +230,34 @@ def test_stack():
     assert s.shape == (3, 4, 6)
     assert s.chunks == ((1, 1, 1), (2, 2), (3, 3))
     assert s.dask[(s.name, 0, 1, 0)] == (getitem, ('A', 1, 0),
-                                          (None, colon, colon))
+                                         (None, colon, colon))
     assert s.dask[(s.name, 2, 1, 0)] == (getitem, ('C', 1, 0),
-                                          (None, colon, colon))
+                                         (None, colon, colon))
     assert same_keys(s, stack([a, b, c], axis=0))
 
     s2 = stack([a, b, c], axis=1)
     assert s2.shape == (4, 3, 6)
     assert s2.chunks == ((2, 2), (1, 1, 1), (3, 3))
     assert s2.dask[(s2.name, 0, 1, 0)] == (getitem, ('B', 0, 0),
-                                            (colon, None, colon))
+                                           (colon, None, colon))
     assert s2.dask[(s2.name, 1, 1, 0)] == (getitem, ('B', 1, 0),
-                                            (colon, None, colon))
+                                           (colon, None, colon))
     assert same_keys(s2, stack([a, b, c], axis=1))
 
     s2 = stack([a, b, c], axis=2)
     assert s2.shape == (4, 6, 3)
     assert s2.chunks == ((2, 2), (3, 3), (1, 1, 1))
     assert s2.dask[(s2.name, 0, 1, 0)] == (getitem, ('A', 0, 1),
-                                            (colon, colon, None))
+                                           (colon, colon, None))
     assert s2.dask[(s2.name, 1, 1, 2)] == (getitem, ('C', 1, 1),
-                                            (colon, colon, None))
+                                           (colon, colon, None))
     assert same_keys(s2, stack([a, b, c], axis=2))
 
-    assert raises(ValueError, lambda: stack([a, b, c], axis=3))
+    pytest.raises(ValueError, lambda: stack([a, b, c], axis=3))
 
     assert set(b.dask.keys()).issubset(s2.dask.keys())
 
-    assert stack([a, b, c], axis=-1).chunks == \
-            stack([a, b, c], axis=2).chunks
+    assert stack([a, b, c], axis=-1).chunks == stack([a, b, c], axis=2).chunks
 
 
 def test_short_stack():
@@ -266,10 +276,23 @@ def test_stack_scalars():
     assert s.compute().tolist() == [np.arange(4).mean(), np.arange(4).sum()]
 
 
+@pytest.mark.skipif(LooseVersion(np.__version__) < '1.10.0',
+                    reason="NumPy doesn't yet support stack")
+def test_stack_rechunk():
+    x = da.random.random(10, chunks=5)
+    y = da.random.random(10, chunks=4)
+
+    z = da.stack([x, y], axis=0)
+    assert z.shape == (2, 10)
+    assert z.chunks == ((1, 1), (4, 1, 3, 2))
+
+    assert_eq(z, np.stack([x.compute(), y.compute()], axis=0))
+
+
 def test_concatenate():
     a, b, c = [Array(getem(name, chunks=(2, 3), shape=(4, 6)),
                      name, shape=(4, 6), chunks=(2, 3))
-                for name in 'ABC']
+               for name in 'ABC']
 
     x = concatenate([a, b, c], axis=0)
 
@@ -289,10 +312,25 @@ def test_concatenate():
 
     assert set(b.dask.keys()).issubset(y.dask.keys())
 
-    assert concatenate([a, b, c], axis=-1).chunks == \
-            concatenate([a, b, c], axis=1).chunks
+    assert (concatenate([a, b, c], axis=-1).chunks ==
+            concatenate([a, b, c], axis=1).chunks)
 
-    assert raises(ValueError, lambda: concatenate([a, b, c], axis=2))
+    pytest.raises(ValueError, lambda: concatenate([a, b, c], axis=2))
+
+
+def test_concatenate_rechunk():
+    x = da.random.random((6, 6), chunks=(3, 3))
+    y = da.random.random((6, 6), chunks=(2, 2))
+
+    z = da.concatenate([x, y], axis=0)
+    assert z.shape == (12, 6)
+    assert z.chunks == ((3, 3, 2, 2, 2), (2, 1, 1, 2))
+    assert_eq(z, np.concatenate([x.compute(), y.compute()], axis=0))
+
+    z = da.concatenate([x, y], axis=1)
+    assert z.shape == (6, 12)
+    assert z.chunks == ((2, 1, 1, 2), (3, 3, 2, 2, 2))
+    assert_eq(z, np.concatenate([x.compute(), y.compute()], axis=1))
 
 
 def test_concatenate_fixlen_strings():
@@ -346,7 +384,7 @@ def test_take():
 
     assert_eq(np.take(x, 3, axis=0), take(a, 3, axis=0))
     assert_eq(np.take(x, [3, 4, 5], axis=-1), take(a, [3, 4, 5], axis=-1))
-    assert raises(ValueError, lambda: take(a, 3, axis=2))
+    pytest.raises(ValueError, lambda: take(a, 3, axis=2))
     assert same_keys(take(a, [3, 4, 5], axis=-1), take(a, [3, 4, 5], axis=-1))
 
 
@@ -406,8 +444,8 @@ def test_broadcast_shapes():
     assert (3, 4, 5) == broadcast_shapes((3, 4, 5), (4, 1), ())
     assert (3, 4) == broadcast_shapes((3, 1), (1, 4), (4,))
     assert (5, 6, 7, 3, 4) == broadcast_shapes((3, 1), (), (5, 6, 7, 1, 4))
-    assert raises(ValueError, lambda: broadcast_shapes((3,), (3, 4)))
-    assert raises(ValueError, lambda: broadcast_shapes((2, 3), (2, 3, 1)))
+    pytest.raises(ValueError, lambda: broadcast_shapes((3,), (3, 4)))
+    pytest.raises(ValueError, lambda: broadcast_shapes((2, 3), (2, 3, 1)))
 
 
 def test_elemwise_on_scalars():
@@ -440,8 +478,8 @@ def test_elemwise_with_ndarrays():
     assert_eq(a + y, x + y)
     assert_eq(y + a, x + y)
     # Error on shape mismatch
-    assert raises(ValueError, lambda: a + y.T)
-    assert raises(ValueError, lambda: a + np.arange(2))
+    pytest.raises(ValueError, lambda: a + y.T)
+    pytest.raises(ValueError, lambda: a + np.arange(2))
 
 
 def test_elemwise_differently_chunked():
@@ -469,7 +507,7 @@ def test_operators():
     expr = (3 / a * b)**2 > 5
     assert_eq(expr, (3 / x * y)**2 > 5)
 
-    c = exp(a)
+    c = da.exp(a)
     assert_eq(c, np.exp(x))
 
     assert_eq(abs(-a), a)
@@ -588,9 +626,9 @@ def test_coarsen():
     d = from_array(x, chunks=(4, 8))
 
     assert_eq(chunk.coarsen(np.sum, x, {0: 2, 1: 4}),
-                    coarsen(np.sum, d, {0: 2, 1: 4}))
+              coarsen(np.sum, d, {0: 2, 1: 4}))
     assert_eq(chunk.coarsen(np.sum, x, {0: 2, 1: 4}),
-                    coarsen(da.sum, d, {0: 2, 1: 4}))
+              coarsen(da.sum, d, {0: 2, 1: 4}))
 
 
 def test_coarsen_with_excess():
@@ -612,20 +650,20 @@ def test_insert():
     assert_eq(np.insert(x, 5, -1, axis=1), insert(a, 5, -1, axis=1))
     assert_eq(np.insert(x, -1, -1, axis=-2), insert(a, -1, -1, axis=-2))
     assert_eq(np.insert(x, [2, 3, 3], -1, axis=1),
-                 insert(a, [2, 3, 3], -1, axis=1))
+              insert(a, [2, 3, 3], -1, axis=1))
     assert_eq(np.insert(x, [2, 3, 8, 8, -2, -2], -1, axis=0),
-                 insert(a, [2, 3, 8, 8, -2, -2], -1, axis=0))
+              insert(a, [2, 3, 8, 8, -2, -2], -1, axis=0))
     assert_eq(np.insert(x, slice(1, 4), -1, axis=1),
-                 insert(a, slice(1, 4), -1, axis=1))
+              insert(a, slice(1, 4), -1, axis=1))
     assert_eq(np.insert(x, [2] * 3 + [5] * 2, y, axis=0),
-                 insert(a, [2] * 3 + [5] * 2, b, axis=0))
+              insert(a, [2] * 3 + [5] * 2, b, axis=0))
     assert_eq(np.insert(x, 0, y[0], axis=1),
-                 insert(a, 0, b[0], axis=1))
-    assert raises(NotImplementedError, lambda: insert(a, [4, 2], -1, axis=0))
-    assert raises(IndexError, lambda: insert(a, [3], -1, axis=2))
-    assert raises(IndexError, lambda: insert(a, [3], -1, axis=-3))
+              insert(a, 0, b[0], axis=1))
+    pytest.raises(NotImplementedError, lambda: insert(a, [4, 2], -1, axis=0))
+    pytest.raises(IndexError, lambda: insert(a, [3], -1, axis=2))
+    pytest.raises(IndexError, lambda: insert(a, [3], -1, axis=-3))
     assert same_keys(insert(a, [2, 3, 8, 8, -2, -2], -1, axis=0),
-                    insert(a, [2, 3, 8, 8, -2, -2], -1, axis=0))
+                     insert(a, [2, 3, 8, 8, -2, -2], -1, axis=0))
 
 
 def test_multi_insert():
@@ -641,10 +679,10 @@ def test_broadcast_to():
 
     for shape in [(5, 4, 6), (2, 5, 1, 6), (3, 4, 5, 4, 6)]:
         assert_eq(chunk.broadcast_to(x, shape),
-                        broadcast_to(a, shape))
+                  broadcast_to(a, shape))
 
-    assert raises(ValueError, lambda: broadcast_to(a, (2, 1, 6)))
-    assert raises(ValueError, lambda: broadcast_to(a, (3,)))
+    pytest.raises(ValueError, lambda: broadcast_to(a, (2, 1, 6)))
+    pytest.raises(ValueError, lambda: broadcast_to(a, (3,)))
 
 
 def test_ravel():
@@ -734,7 +772,7 @@ def test_reshape_unknown_dimensions():
             a = from_array(x, 24)
             assert_eq(x.reshape(new_shape), a.reshape(new_shape))
 
-    assert raises(ValueError, lambda: reshape(a, (-1, -1)))
+    pytest.raises(ValueError, lambda: reshape(a, (-1, -1)))
 
 
 def test_full():
@@ -763,7 +801,7 @@ def test_map_blocks():
 
     d = from_array(x, chunks=(8, 8))
     e = d.map_blocks(lambda x: x[::2, ::2], chunks=((4, 4, 2), (4, 4, 2)),
-            dtype=d.dtype)
+                     dtype=d.dtype)
 
     assert_eq(e, x[::2, ::2])
 
@@ -772,8 +810,8 @@ def test_map_blocks2():
     x = np.arange(10, dtype='i8')
     d = from_array(x, chunks=(2,))
 
-    def func(block, block_id=None):
-        return np.ones_like(block) * sum(block_id)
+    def func(block, block_id=None, c=0):
+        return np.ones_like(block) * sum(block_id) + c
 
     out = d.map_blocks(func, dtype='i8')
     expected = np.array([0, 0, 1, 1, 2, 2, 3, 3, 4, 4], dtype='i8')
@@ -781,6 +819,12 @@ def test_map_blocks2():
     assert_eq(out, expected)
     assert same_keys(d.map_blocks(func, dtype='i8'), out)
 
+    out = d.map_blocks(func, dtype='i8', c=1)
+    expected = expected + 1
+
+    assert_eq(out, expected)
+    assert same_keys(d.map_blocks(func, dtype='i8', c=1), out)
+
 
 def test_map_blocks_with_constants():
     d = da.arange(10, chunks=3)
@@ -812,7 +856,7 @@ def test_fromfunction():
 
 def test_from_function_requires_block_args():
     x = np.arange(10)
-    assert raises(Exception, lambda: from_array(x))
+    pytest.raises(Exception, lambda: from_array(x))
 
 
 def test_repr():
@@ -854,7 +898,7 @@ def test_dtype():
 
 def test_blockdims_from_blockshape():
     assert blockdims_from_blockshape((10, 10), (4, 3)) == ((4, 4, 2), (3, 3, 3, 1))
-    assert raises(TypeError, lambda: blockdims_from_blockshape((10,), None))
+    pytest.raises(TypeError, lambda: blockdims_from_blockshape((10,), None))
     assert blockdims_from_blockshape((1e2, 3), [1e1, 3]) == ((10, ) * 10, (3, ))
     assert blockdims_from_blockshape((np.int8(10), ), (5, )) == ((5, 5), )
 
@@ -879,9 +923,9 @@ def test_store():
     assert (at == 2).all()
     assert (bt == 3).all()
 
-    assert raises(ValueError, lambda: store([a], [at, bt]))
-    assert raises(ValueError, lambda: store(at, at))
-    assert raises(ValueError, lambda: store([at, bt], [at, bt]))
+    pytest.raises(ValueError, lambda: store([a], [at, bt]))
+    pytest.raises(ValueError, lambda: store(at, at))
+    pytest.raises(ValueError, lambda: store([at, bt], [at, bt]))
 
 
 def test_store_compute_false():
@@ -1077,14 +1121,31 @@ def test_dtype_complex():
 
 
 def test_astype():
-    x = np.ones(5, dtype='f4')
-    d = da.from_array(x, chunks=(2,))
+    x = np.ones((5, 5), dtype='f8')
+    d = da.from_array(x, chunks=(2,2))
 
     assert d.astype('i8')._dtype == 'i8'
     assert_eq(d.astype('i8'), x.astype('i8'))
     assert same_keys(d.astype('i8'), d.astype('i8'))
 
-    assert d.astype(d.dtype) is d
+    with pytest.raises(TypeError):
+        d.astype('i8', casting='safe')
+
+    with pytest.raises(TypeError):
+        d.astype('i8', not_a_real_kwarg='foo')
+
+    # smoketest with kwargs
+    assert_eq(d.astype('i8', copy=False), x.astype('i8', copy=False))
+
+    # Check it's a noop
+    assert d.astype('f8') is d
+
+    # Check arrays with unknown dtype are not noops. Since `dtype('f8') == None`
+    # is True, comparing dtypes directly can lead to bugs
+    d._dtype = None
+    d2 = d.astype('f8')
+    assert d2 is not d
+    assert_eq(d2, x.astype('f8'))
 
 
 def test_arithmetic():
@@ -1234,18 +1295,6 @@ def test_arithmetic():
     assert_eq(da.around(a, -1), np.around(x, -1))
 
 
-def test_clip():
-    x = np.random.normal(0, 10, size=(10, 10))
-    d = da.from_array(x, chunks=(3, 4))
-
-    assert_eq(x.clip(5), d.clip(5))
-    assert_eq(x.clip(1, 5), d.clip(1, 5))
-    assert_eq(x.clip(min=5), d.clip(min=5))
-    assert_eq(x.clip(max=5), d.clip(max=5))
-    assert_eq(x.clip(max=1, min=5), d.clip(max=1, min=5))
-    assert_eq(x.clip(min=1, max=5), d.clip(min=1, max=5))
-
-
 def test_elemwise_consistent_names():
     a = da.from_array(np.arange(5, dtype='f4'), chunks=(2,))
     b = da.from_array(np.arange(5, dtype='f4'), chunks=(2,))
@@ -1472,7 +1521,7 @@ def test_histogram_return_type():
     # Check if return type is same as hist
     bins = np.arange(0, 11, 1, dtype='i4')
     assert_eq(da.histogram(v * 10, bins=bins)[0],
-       np.histogram(v * 10, bins=bins)[0])
+              np.histogram(v * 10, bins=bins)[0])
 
 
 def test_histogram_extra_args_and_shapes():
@@ -1485,24 +1534,24 @@ def test_histogram_extra_args_and_shapes():
     for v, bins, w in data:
         # density
         assert_eq(da.histogram(v, bins=bins, normed=True)[0],
-           np.histogram(v, bins=bins, normed=True)[0])
+                  np.histogram(v, bins=bins, normed=True)[0])
 
         # normed
         assert_eq(da.histogram(v, bins=bins, density=True)[0],
-           np.histogram(v, bins=bins, density=True)[0])
+                  np.histogram(v, bins=bins, density=True)[0])
 
         # weights
         assert_eq(da.histogram(v, bins=bins, weights=w)[0],
-           np.histogram(v, bins=bins, weights=w)[0])
+                  np.histogram(v, bins=bins, weights=w)[0])
 
         assert_eq(da.histogram(v, bins=bins, weights=w, density=True)[0],
-           da.histogram(v, bins=bins, weights=w, density=True)[0])
+                  da.histogram(v, bins=bins, weights=w, density=True)[0])
 
 
 def test_concatenate3_2():
     x = np.array([1, 2])
     assert_eq(concatenate3([x, x, x]),
-       np.array([1, 2, 1, 2, 1, 2]))
+              np.array([1, 2, 1, 2, 1, 2]))
 
     x = np.array([[1, 2]])
     assert (concatenate3([[x, x, x], [x, x, x]]) ==
@@ -1515,29 +1564,27 @@ def test_concatenate3_2():
                       [1, 2, 1, 2]])).all()
 
     x = np.arange(12).reshape((2, 2, 3))
-    assert_eq(concatenate3([[[x, x, x],
-                      [x, x, x]],
-                     [[x, x, x],
-                      [x, x, x]]]),
-       np.array([[[ 0,  1,  2,  0,  1,  2,  0,  1,  2],
-                  [ 3,  4,  5,  3,  4,  5,  3,  4,  5],
-                  [ 0,  1,  2,  0,  1,  2,  0,  1,  2],
-                  [ 3,  4,  5,  3,  4,  5,  3,  4,  5]],
+    assert_eq(concatenate3([[[x, x, x], [x, x, x]],
+                           [[x, x, x], [x, x, x]]]),
+              np.array([[[ 0,  1,  2,  0,  1,  2,  0,  1,  2],
+                         [ 3,  4,  5,  3,  4,  5,  3,  4,  5],
+                         [ 0,  1,  2,  0,  1,  2,  0,  1,  2],
+                         [ 3,  4,  5,  3,  4,  5,  3,  4,  5]],
 
-                 [[ 6,  7,  8,  6,  7,  8,  6,  7,  8],
-                  [ 9, 10, 11,  9, 10, 11,  9, 10, 11],
-                  [ 6,  7,  8,  6,  7,  8,  6,  7,  8],
-                  [ 9, 10, 11,  9, 10, 11,  9, 10, 11]],
+                        [[ 6,  7,  8,  6,  7,  8,  6,  7,  8],
+                         [ 9, 10, 11,  9, 10, 11,  9, 10, 11],
+                         [ 6,  7,  8,  6,  7,  8,  6,  7,  8],
+                         [ 9, 10, 11,  9, 10, 11,  9, 10, 11]],
 
-                 [[ 0,  1,  2,  0,  1,  2,  0,  1,  2],
-                  [ 3,  4,  5,  3,  4,  5,  3,  4,  5],
-                  [ 0,  1,  2,  0,  1,  2,  0,  1,  2],
-                  [ 3,  4,  5,  3,  4,  5,  3,  4,  5]],
+                        [[ 0,  1,  2,  0,  1,  2,  0,  1,  2],
+                         [ 3,  4,  5,  3,  4,  5,  3,  4,  5],
+                         [ 0,  1,  2,  0,  1,  2,  0,  1,  2],
+                         [ 3,  4,  5,  3,  4,  5,  3,  4,  5]],
 
-                 [[ 6,  7,  8,  6,  7,  8,  6,  7,  8],
-                  [ 9, 10, 11,  9, 10, 11,  9, 10, 11],
-                  [ 6,  7,  8,  6,  7,  8,  6,  7,  8],
-                  [ 9, 10, 11,  9, 10, 11,  9, 10, 11]]]))
+                        [[ 6,  7,  8,  6,  7,  8,  6,  7,  8],
+                         [ 9, 10, 11,  9, 10, 11,  9, 10, 11],
+                         [ 6,  7,  8,  6,  7,  8,  6,  7,  8],
+                         [ 9, 10, 11,  9, 10, 11,  9, 10, 11]]]))
 
 
 def test_map_blocks3():
@@ -1603,7 +1650,7 @@ def test_raise_on_no_chunks():
     except ValueError as e:
         assert "dask.pydata.org" in str(e)
 
-    assert raises(ValueError, lambda: da.ones(6))
+    pytest.raises(ValueError, lambda: da.ones(6))
 
 
 def test_chunks_is_immutable():
@@ -1665,20 +1712,18 @@ def test_point_slicing_with_full_slice():
     x = np.arange(4 * 5 * 6 * 7).reshape((4, 5, 6, 7))
     d = da.from_array(x, chunks=(2, 3, 3, 4))
 
-    inds = [
-            [[1, 2, 3], None, [3, 2, 1], [5, 3, 4]],
+    inds = [[[1, 2, 3], None, [3, 2, 1], [5, 3, 4]],
             [[1, 2, 3], None, [4, 3, 2], None],
             [[1, 2, 3], [3, 2, 1]],
             [[1, 2, 3], [3, 2, 1], [3, 2, 1], [5, 3, 4]],
             [[], [], [], None],
             [np.array([1, 2, 3]), None, np.array([4, 3, 2]), None],
             [None, None, [1, 2, 3], [4, 3, 2]],
-            [None, [0, 2, 3], None, [0, 3, 2]],
-            ]
+            [None, [0, 2, 3], None, [0, 3, 2]]]
 
     for ind in inds:
         slc = [i if isinstance(i, (np.ndarray, list)) else slice(None, None)
-                for i in ind]
+               for i in ind]
         result = d.vindex[tuple(slc)]
 
         # Rotate the expected result accordingly
@@ -1704,11 +1749,11 @@ def test_slice_with_floats():
 
 def test_vindex_errors():
     d = da.ones((5, 5, 5), chunks=(3, 3, 3))
-    assert raises(IndexError, lambda: d.vindex[0])
-    assert raises(IndexError, lambda: d.vindex[[1, 2, 3]])
-    assert raises(IndexError, lambda: d.vindex[[1, 2, 3], [1, 2, 3], 0])
-    assert raises(IndexError, lambda: d.vindex[[1], [1, 2, 3]])
-    assert raises(IndexError, lambda: d.vindex[[1, 2, 3], [[1], [2], [3]]])
+    pytest.raises(IndexError, lambda: d.vindex[0])
+    pytest.raises(IndexError, lambda: d.vindex[[1, 2, 3]])
+    pytest.raises(IndexError, lambda: d.vindex[[1, 2, 3], [1, 2, 3], 0])
+    pytest.raises(IndexError, lambda: d.vindex[[1], [1, 2, 3]])
+    pytest.raises(IndexError, lambda: d.vindex[[1, 2, 3], [[1], [2], [3]]])
 
 
 def test_vindex_merge():
@@ -1730,7 +1775,7 @@ def test_array():
     x = np.ones(5, dtype='i4')
     d = da.ones(5, chunks=3, dtype='i4')
     assert_eq(da.array(d, ndmin=3, dtype='i8'),
-       np.array(x, ndmin=3, dtype='i8'))
+              np.array(x, ndmin=3, dtype='i8'))
 
 
 def test_cov():
@@ -1749,7 +1794,7 @@ def test_cov():
     assert_eq(da.cov(d, e), np.cov(x, y))
     assert_eq(da.cov(e, d), np.cov(y, x))
 
-    assert raises(ValueError, lambda: da.cov(d, ddof=1.5))
+    pytest.raises(ValueError, lambda: da.cov(d, ddof=1.5))
 
 
 def test_corrcoef():
@@ -1848,15 +1893,38 @@ def test_map_blocks_with_changed_dimension():
 
     e = d.map_blocks(lambda b: b.sum(axis=0), chunks=(4,), drop_axis=0,
                      dtype=d.dtype)
-    assert e.ndim == 1
     assert e.chunks == ((4, 4),)
     assert_eq(e, x.sum(axis=0))
 
+    # Provided chunks have wrong shape
+    with pytest.raises(ValueError):
+        d.map_blocks(lambda b: b.sum(axis=0), chunks=(7, 4), drop_axis=0)
+
+    with pytest.raises(ValueError):
+        d.map_blocks(lambda b: b.sum(axis=0), chunks=((4, 4, 4),), drop_axis=0)
+
+    # Can't drop axis with more than 1 block
+    with pytest.raises(ValueError):
+        d.map_blocks(lambda b: b.sum(axis=1), drop_axis=1, dtype=d.dtype)
+
+    # Can't use both drop_axis and new_axis
+    with pytest.raises(ValueError):
+        d.map_blocks(lambda b: b, drop_axis=1, new_axis=1)
+
+    d = da.from_array(x, chunks=(4, 8))
+    e = d.map_blocks(lambda b: b.sum(axis=1), drop_axis=1, dtype=d.dtype)
+    assert e.chunks == ((4, 3),)
+    assert_eq(e, x.sum(axis=1))
+
     x = np.arange(64).reshape((8, 8))
     d = da.from_array(x, chunks=(4, 4))
     e = d.map_blocks(lambda b: b[None, :, :, None],
                      chunks=(1, 4, 4, 1), new_axis=[0, 3], dtype=d.dtype)
-    assert e.ndim == 4
+    assert e.chunks == ((1,), (4, 4), (4, 4), (1,))
+    assert_eq(e, x[None, :, :, None])
+
+    e = d.map_blocks(lambda b: b[None, :, :, None],
+                     new_axis=[0, 3], dtype=d.dtype)
     assert e.chunks == ((1,), (4, 4), (4, 4), (1,))
     assert_eq(e, x[None, :, :, None])
 
@@ -2044,11 +2112,11 @@ def test_tril_triu():
 def test_tril_triu_errors():
     A = np.random.random_integers(0, 10, (10, 10, 10))
     dA = da.from_array(A, chunks=(5, 5, 5))
-    assert raises(ValueError, lambda: da.triu(dA))
+    pytest.raises(ValueError, lambda: da.triu(dA))
 
     A = np.random.random_integers(0, 10, (30, 35))
     dA = da.from_array(A, chunks=(5, 5))
-    assert raises(NotImplementedError, lambda: da.triu(dA))
+    pytest.raises(NotImplementedError, lambda: da.triu(dA))
 
 
 def test_atop_names():
@@ -2094,6 +2162,34 @@ def test_atop_kwargs():
     assert_eq(y, np.ones(5) + 10)
 
 
+def test_atop_chunks():
+    x = da.ones((5, 5), chunks=((2, 1, 2), (3, 2)))
+
+    def double(a, axis=0):
+        return np.concatenate([a, a], axis=axis)
+
+    y = atop(double, 'ij', x, 'ij',
+             adjust_chunks={'i': lambda n: 2 * n}, axis=0, dtype=x.dtype)
+    assert y.chunks == ((4, 2, 4), (3, 2))
+    assert_eq(y, np.ones((10, 5)))
+
+    y = atop(double, 'ij', x, 'ij',
+             adjust_chunks={'j': lambda n: 2 * n}, axis=1, dtype=x.dtype)
+    assert y.chunks == ((2, 1, 2), (6, 4))
+    assert_eq(y, np.ones((5, 10)))
+
+    x = da.ones((10, 10), chunks=(5, 5))
+    y = atop(double, 'ij', x, 'ij', axis=0,
+             adjust_chunks={'i': 10}, dtype=x.dtype)
+    assert y.chunks == ((10, 10), (5, 5))
+    assert_eq(y, np.ones((20, 10)))
+
+    y = atop(double, 'ij', x, 'ij', axis=0,
+             adjust_chunks={'i': (10, 10)}, dtype=x.dtype)
+    assert y.chunks == ((10, 10), (5, 5))
+    assert_eq(y, np.ones((20, 10)))
+
+
 def test_from_delayed():
     v = delayed(np.ones)((5, 3))
     x = from_delayed(v, shape=(5, 3), dtype=np.ones(0).dtype)
@@ -2217,7 +2313,6 @@ def test_atop_concatenate():
     z = atop(add, 'ij', y, 'ij', y, 'ij', concatenate=True, dtype=x._dtype)
     assert_eq(z, np.ones((4, 4)) * 2)
 
-
     def f(a, b, c):
         assert isinstance(a, np.ndarray)
         assert isinstance(b, np.ndarray)
@@ -2232,3 +2327,124 @@ def test_atop_concatenate():
     z = atop(f, 'j', x, 'ijk', y, 'ki', y, 'ij', concatenate=True,
              dtype=x._dtype)
     assert_eq(z, np.ones(10))
+
+
+def test_common_blockdim():
+    assert common_blockdim([(5,), (5,)]) == (5,)
+    assert common_blockdim([(5,), (2, 3,)]) == (2, 3)
+    assert common_blockdim([(5, 5), (2, 3, 5)]) == (2, 3, 5)
+    assert common_blockdim([(5, 5), (2, 3, 5)]) == (2, 3, 5)
+    assert common_blockdim([(5, 2, 3), (2, 3, 5)]) == (2, 3, 2, 3)
+
+    assert common_blockdim([(1, 2), (2, 1)]) == (1, 1, 1)
+    assert common_blockdim([(1, 2, 2), (2, 1, 2), (2, 2, 1)]) == (1, 1, 1, 1, 1)
+
+
+def test_uneven_chunks_that_fit_neatly():
+    x = da.arange(10, chunks=((5, 5),))
+    y = da.ones(10, chunks=((5, 2, 3),))
+
+    assert_eq(x + y, np.arange(10) + np.ones(10))
+
+    z = x + y
+    assert z.chunks == ((5, 2, 3),)
+
+
+def test_elemwise_uneven_chunks():
+    x = da.arange(10, chunks=((4, 6),))
+    y = da.ones(10, chunks=((6, 4),))
+
+    assert_eq(x + y, np.arange(10) + np.ones(10))
+
+    z = x + y
+    assert z.chunks == ((4, 2, 4),)
+
+    x = da.random.random((10, 10), chunks=((4, 6), (5, 2, 3)))
+    y = da.random.random((4, 10, 10), chunks=((2, 2), (6, 4), (2, 3, 5)))
+
+    z = x + y
+    assert_eq(x + y, x.compute() + y.compute())
+    assert z.chunks == ((2, 2), (4, 2, 4), (2, 3, 2, 3))
+
+
+def test_uneven_chunks_atop():
+    x = da.random.random((10, 10), chunks=((2, 3, 2, 3), (5, 5)))
+    y = da.random.random((10, 10), chunks=((4, 4, 2), (4, 2, 4)))
+    z = atop(np.dot, 'ik', x, 'ij', y, 'jk', dtype=x._dtype, concatenate=True)
+    assert z.chunks == (x.chunks[0], y.chunks[1])
+
+    assert_eq(z, x.compute().dot(y))
+
+
+def test_warn_bad_rechunking():
+    x = da.ones((20, 20), chunks=(20, 1))
+    y = da.ones((20, 20), chunks=(1, 20))
+
+    with warnings.catch_warnings(record=True) as record:
+        x + y
+
+    assert record
+    assert '20' in record[0].message.args[0]
+
+
+def test_optimize_fuse_keys():
+    x = da.ones(10, chunks=(5,))
+    y = x + 1
+    z = y + 1
+
+    dsk = z._optimize(z.dask, z._keys())
+    assert not set(y.dask) & set(dsk)
+
+    dsk = z._optimize(z.dask, z._keys(), fuse_keys=y._keys())
+    assert all(k in dsk for k in y._keys())
+
+
+def test_round():
+    x = np.random.random(10)
+    d = da.from_array(x, chunks=4)
+
+    for i in (0, 1, 4, 5):
+        assert_eq(x.round(i), d.round(i))
+
+    assert_eq(d.round(2), da.round(d, 2))
+
+
+def test_repeat():
+    x = np.random.random((10, 11, 13))
+    d = da.from_array(x, chunks=(4, 5, 3))
+
+    repeats = [1, 2, 5]
+    axes = [0, 1, 2]
+
+    for r in repeats:
+        for a in axes:
+            assert_eq(x.repeat(r, axis=a), d.repeat(r, axis=a))
+
+    assert_eq(d.repeat(2, 0), da.repeat(d, 2, 0))
+
+    with pytest.raises(NotImplementedError):
+        da.repeat(d, np.arange(10))
+
+    with pytest.raises(NotImplementedError):
+        da.repeat(d, 2, None)
+
+    with pytest.raises(NotImplementedError):
+        da.repeat(d, 2)
+
+    x = np.arange(5)
+    d = da.arange(5, chunks=(2,))
+
+    assert_eq(x.repeat(3), d.repeat(3))
+
+    for r in [1, 2, 3, 4]:
+        assert all(concat(d.repeat(r).chunks))
+
+
+def test_concatenate_stack_dont_warn():
+    with warnings.catch_warnings(record=True) as record:
+        da.concatenate([da.ones(2, chunks=1)] * 62)
+    assert not record
+
+    with warnings.catch_warnings(record=True) as record:
+        da.stack([da.ones(2, chunks=1)] * 62)
+    assert not record
