@@ -27,14 +27,13 @@ X = da.from_array(x, chunks=(3, 2))
 Y = da.from_array(y, chunks=(3,))
 Z = da.from_array(z, chunks=(2, 2))
 
+
 def test_fit():
     sgd = SGDClassifier()
 
-    sgd = da.learn.fit(sgd, X, Y, get=dask.get, classes=np.array([-1, 1]))
-
-    result = sgd.predict(z)
-    assert result.tolist() == [1, -1, 1, -1]
+    sgd = da.learn.fit(sgd, X, Y, get=dask.get, classes=np.array([-1, 0, 1]))
 
+    sol = sgd.predict(z)
     result = da.learn.predict(sgd, Z)
     assert result.chunks == ((2, 2),)
-    assert result.compute(get=dask.get).tolist() == [1, -1, 1, -1]
+    assert result.compute(get=dask.get).tolist() == sol.tolist()
