@@ -1,5 +1,6 @@
 import numpy as np
 import pandas as pd
+import pandas.util.testing as tm
 import dask.dataframe as dd
 from dask.dataframe.utils import (shard_df_on_index, meta_nonempty, make_meta,
                                   raise_on_meta_error)
@@ -127,6 +128,17 @@ def test_meta_nonempty():
     assert (df3['A'] == s).all()
 
 
+def test_meta_duplicated():
+    df = pd.DataFrame(columns=['A', 'A', 'B'])
+    res = meta_nonempty(df)
+
+    exp = pd.DataFrame([['foo', 'foo', 'foo'],
+                        ['foo', 'foo', 'foo']],
+                       index=['a', 'b'],
+                       columns=['A', 'A', 'B'])
+    tm.assert_frame_equal(res, exp)
+
+
 def test_meta_nonempty_index():
     idx = pd.RangeIndex(1, name='foo')
     res = meta_nonempty(idx)
