@@ -126,12 +126,14 @@ def test_cache_profiler():
     assert cprof.results == []
 
     tics = [0]
+
     def nbytes(res):
         tics[0] += 1
         return tics[0]
 
     with CacheProfiler(nbytes) as cprof:
         get(dsk2, 'c')
+
     results = cprof.results
     assert tics[-1] == len(results)
     assert tics[-1] == results[-1].metric
@@ -164,13 +166,15 @@ def test_pprint_task():
 
     assert len(pprint_task((sum, list(keys) * 100), keys)) < 100
     assert pprint_task((sum, list(keys) * 100), keys) == 'sum([_, _, _, ...])'
-    assert pprint_task((sum, [1, 2, (sum, ['a', 4]), 5, 6] * 100), keys) == \
-            'sum([*, *, sum([_, *]), ...])'
+    assert (pprint_task((sum, [1, 2, (sum, ['a', 4]), 5, 6] * 100), keys) ==
+            'sum([*, *, sum([_, *]), ...])')
     assert pprint_task((sum, [1, 2, (sum, ['a', (sum, [1, 2, 3])]), 5, 6]),
                        keys) == 'sum([*, *, sum([_, sum(...)]), ...])'
+
     # With kwargs
     def foo(w, x, y=(), z=3):
         return w + x + sum(y) + z
+
     task = (apply, foo, (tuple, ['a', 'b']),
             (dict, [['y', ['a', 'b']], ['z', 'c']]))
     assert pprint_task(task, keys) == 'foo(_, _, y=[_, _], z=_)'
