@@ -34,18 +34,26 @@ def cull(dsk, keys):
     """
     if not isinstance(keys, (list, set)):
         keys = [keys]
-    out = dict()
+    out_keys = []
     seen = set()
     dependencies = dict()
-    stack = list(set(flatten(keys)))
-    while stack:
-        key = stack.pop()
-        out[key] = dsk[key]
-        deps = get_dependencies(dsk, key, as_list=True)  # fuse needs lists
-        dependencies[key] = deps
-        unseen = [d for d in deps if d not in seen]
-        stack.extend(unseen)
-        seen.update(unseen)
+
+    work = list(set(flatten(keys)))
+    while work:
+        new_work = []
+        out_keys += work
+        deps = [(k, get_dependencies(dsk, k, as_list=True))  # fuse needs lists
+                for k in work]
+        dependencies.update(deps)
+        for _, deplist in deps:
+            for d in deplist:
+                if d not in seen:
+                    seen.add(d)
+                    new_work.append(d)
+        work = new_work
+
+    out = {k: dsk[k] for k in out_keys}
+
     return out, dependencies
 
 
@@ -86,8 +94,8 @@ def fuse(dsk, keys=None, dependencies=None):
         keys = set(flatten(keys))
 
     if dependencies is None:
-        dependencies = dict((key, get_dependencies(dsk, key, as_list=True))
-                            for key in dsk)
+        dependencies = {k: get_dependencies(dsk, k, as_list=True)
+                        for k in dsk}
 
     # locate all members of linear chains
     child2parent = {}
@@ -182,7 +190,8 @@ def inline(dsk, keys=None, inline_constants=True, dependencies=None):
     keys = _flat_set(keys)
 
     if dependencies is None:
-        dependencies = dict((k, get_dependencies(dsk, k)) for k in dsk)
+        dependencies = {k: get_dependencies(dsk, k)
+                        for k in dsk}
 
     if inline_constants:
         keys.update(k for k, v in dsk.items() if
@@ -215,7 +224,7 @@ def inline(dsk, keys=None, inline_constants=True, dependencies=None):
 
 
 def inline_functions(dsk, output, fast_functions=None, inline_constants=False,
-        dependencies=None):
+                     dependencies=None):
     """ Inline cheap functions into larger operations
 
     Examples
@@ -245,22 +254,29 @@ def inline_functions(dsk, output, fast_functions=None, inline_constants=False,
     fast_functions = set(fast_functions)
 
     if dependencies is None:
-        dependencies = dict((k, get_dependencies(dsk, k)) for k in dsk)
+        dependencies = {k: get_dependencies(dsk, k)
+                        for k in dsk}
     dependents = reverse_dict(dependencies)
 
     keys = [k for k, v in dsk.items()
-              if istask(v)
-              and functions_of(v).issubset(fast_functions)
-              and dependents[k]
-              and k not in output]
+            if istask(v) and functions_of(v).issubset(fast_functions) and
+            dependents[k] and k not in output
+            ]
+
     if keys:
         dsk = inline(dsk, keys, inline_constants=inline_constants,
-                dependencies=dependencies)
+                     dependencies=dependencies)
         for k in keys:
             del dsk[k]
     return dsk
 
 
+def unwrap_partial(func):
+    while hasattr(func, 'func'):
+        func = func.func
+    return func
+
+
 def functions_of(task):
     """ Set of functions contained within nested task
 
@@ -270,20 +286,23 @@ def functions_of(task):
     >>> functions_of(task)  # doctest: +SKIP
     set([add, mul, inc])
     """
-    if istask(task):
-        args = set.union(*map(functions_of, task[1:])) if task[1:] else set()
-        return set([unwrap_partial(task[0])]) | args
-    if isinstance(task, (list, tuple)):
-        if not task:
-            return set()
-        return set.union(*map(functions_of, task))
-    return set()
+    funcs = set()
 
+    work = [task]
+    sequence_types = {list, tuple}
 
-def unwrap_partial(func):
-    while hasattr(func, 'func'):
-        func = func.func
-    return func
+    while work:
+        new_work = []
+        for task in work:
+            if type(task) in sequence_types:
+                if istask(task):
+                    funcs.add(unwrap_partial(task[0]))
+                    new_work += task[1:]
+                else:
+                    new_work += task
+        work = new_work
+
+    return funcs
 
 
 def dealias(dsk, keys=None, dependencies=None):
@@ -322,7 +341,8 @@ def dealias(dsk, keys=None, dependencies=None):
         keys = set(keys)
 
     if not dependencies:
-        dependencies = dict((k, get_dependencies(dsk, k)) for k in dsk)
+        dependencies = {k: get_dependencies(dsk, k)
+                        for k in dsk}
 
     aliases = set(k for k, task in dsk.items() if
                   ishashable(task) and task in dsk)
@@ -589,4 +609,4 @@ def fuse_getitem(dsk, func, place):
     {'y': (<function load at ...>, 'store', 'part', 'a')}
     """
     return fuse_selections(dsk, getitem, func,
-            lambda a, b: tuple(b[:place]) + (a[2],) + tuple(b[place + 1:]))
+                           lambda a, b: tuple(b[:place]) + (a[2], ) + tuple(b[place + 1:]))
